"""
Tests for ludology.tools.
"""

import pytest

from ludology import Game
from ludology.closet import zero, pm_one, star
from ludology.tools import left_stop, right_stop, left_incentives, right_incentives


@pytest.mark.parametrize(['g', 'v'], [
    (zero, (0, '+')),
    (pm_one, (1, '-')),
    (star, (0, '-')),
])
def test_left_stop_1(g, v):
    """
    Test left stops with adornment.
    """
    assert left_stop(g) == v


@pytest.mark.parametrize(['g', 'v'], [
    (zero, 0),
    (pm_one, 1),
    (star, 0),
])
def test_left_stop_2(g, v):
    """
    Test left stops without adornment.
    """
    assert left_stop(g, adorn=False) == v


@pytest.mark.parametrize(['g', 'v'], [
    (zero, (0, '-')),
    (pm_one, (-1, '+')),
    (star, (0, '+')),
])
def test_right_stop_1(g, v):
    """
    Test right stops with adornment.
    """
    assert right_stop(g) == v


@pytest.mark.parametrize(['g', 'v'], [
    (zero, 0),
    (pm_one, -1),
    (star, 0),
])
def test_right_stop_2(g, v):
    """
    Test right stops without adornment.
    """
    assert right_stop(g, adorn=False) == v


def test_left_incentives():
    """
    Test for the incentives of a known example.
    """
    g = Game({0}, {Game({0}, {-2})})
    li = {Game({Game({2}, {0})}, {0})}
    assert left_incentives(g) == li


def test_right_incentives():
    """
    Test for the incentives of a known example.
    """
    g = Game({0}, {Game({0}, {-2})})
    ri = {Game({Game({2}, {0}), 2}, {0})}
    assert right_incentives(g) == ri



"""
The Surreal numbers are the subset of Games which are isomorophic to the hyperreals.
"""

from fractions import Fraction
from functools import lru_cache, total_ordering
import numbers

from .game import Game


@total_ordering
class Surreal(Game, numbers.Number):
    """
    The Surreal numbers are the subset of games such that G_L <= G_R, and that is
    true for all options as well. The Surreal numbers are isomorphic to the hyperreals.
    All Surreal numbers with finite birthdays correspond to the dyadic rationals, that is,
    rational numbers whose denominator is a power of 2.
    """
    def __init__(G, value):
        """
        Construct the Surreal with given value.

        Parameters
        ----------
        value : int, float, Fraction
            The value for the Surreal number.
        """
        G._n = Fraction(value)
        if G._n.denominator > 1024:
            msg = f"{G._n} would result in a very deep game tree."
            raise ValueError(msg)
        elif G._n.denominator == 1:
            if G._n == 0:
                G._left = set()
                G._right = set()
            elif G._n > 0:
                G._left = {Surreal(value - 1)}
                G._right = set()
            else:
                G._left = set()
                G._right = {Surreal(value + 1)}
        else:
            G._left = {Surreal(G._n - 1/G._n.denominator)}
            G._right = {Surreal(G._n + 1/G._n.denominator)}

    @property
    def is_number(G):
        """
        Whether the Surreal number is a number or not. By definition, this is True.

        Returns
        -------
        number : bool
            Whether this number is a number (True).
        """
        return True

    @property
    def is_numberish(G):
        """
        A Game is numberish if it is infinitesimally close to a number. All Numbers are numberish.

        Returns
        -------
        numberish : bool
            Whether G is numberish or not (it is).
        """
        return True

    @property
    def is_impartial(G):
        """
        Whether the Surreal number is impartial or not. Only the Surreal 0 is impartial.

        Returns
        -------
        impartial : bool
            Whether this number is impartial.
        """
        return G._n == 0

    @property
    def is_dicotic(G):
        """
        Whether the Surreal number is dicotic or not. Only the Surreal 0 is dicotic.

        Returns
        -------
        dicotic : bool
            Whether this number is dicotic.
        """
        return G._n == 0

    @property
    def is_infinitesimal(G):
        """
        A game is infinitesimal if it is non-zero and smaller than any positive number and greater
        than any negative number. Equivalently, it's left and right stops are both zero. Note, this
        does not imply that an infinitesimal can not be positive (> 0) or negative (< 0).

        Returns
        -------
        infinitesimal : bool
            Whether the Surreal is infinitesimal or not (it is not).
        """
        return False

    @property
    def is_switch(G):
        """
        Whether the Surreal number is a switch or not. By definition, this is not the case.

        Returns
        -------
        switch : bool
            Whether this number is a switch (it is not).
        """
        return False

    def __hash__(G):
        """
        Define the hash of a Surreal as the hash of its value.

        Returns
        -------
        hash : str
            The hash of G.
        """
        return hash(G._n)

    @lru_cache(maxsize=None)
    def __ge__(G, H):
        """
        Test if one Surreal is greather than or equal to another.

        Parameters
        ----------
        H : Surreal, Game
            The second Surreal/Game.

        Returns
        ge : bool
            Whether G >= H or not.
        """
        if isinstance(H, Surreal):
            return G._n >= H._n
        elif isinstance(H, Game):
            return super().__ge__(H)
        else:
            return NotImplemented

    def __eq__(G, H):
        """
        Test if one Surreal is equal to another.

        Parameters
        ----------
        H : Surreal, Game
            The second Surreal/Game.

        Returns
        eq : bool
            Whether G == H or not.
        """
        if isinstance(H, Surreal):
            return Surreal(G._n == H._n)
        elif isinstance(H, Game):
            return super().__eq__(H)
        else:
            return NotImplemented

    @lru_cache(maxsize=None)
    def __add__(G, H):
        """
        The sum of two Surreals.

        Parameters
        ----------
        H : Surreal, Game
            The second Surreal/Game.

        Returns
        -------
        sum : Surreal
            The sum of G and H.
        """
        if isinstance(H, Surreal):
            return Surreal(G._n + H._n)
        elif isinstance(H, Game):
            return super().__add__(H)
        else:
            return NotImplemented

    @lru_cache(maxsize=None)
    def __mul__(G, H):
        """
        The product of two Surreals.

        Parameters
        ----------
        H : Surreal, Game
            The second Surreal/Game.

        Returns
        -------
        prod : Surreal
            The product of G and H.
        """
        if isinstance(H, Surreal):
            return Surreal(G._n * H._n)
        elif isinstance(H, Game):
            return super().__mul__(H)
        else:
            return NotImplemented


    @lru_cache(maxsize=None)
    def __truediv__(G, H):
        """
        The quotient of a Surreal by another.

        Parameters
        ----------
        H : Surreal, Game
            The second Surreal/Game.

        Returns
        -------
        div : Surreal
            The quotient of G by H.
        """
        if isinstance(H, Surreal):
            return Surreal(G._n / H._n)
        elif isinstance(H, Game):
            return super().__truediv__(H)
        else:
            return NotImplemented


    @lru_cache(maxsize=None)
    def _invert(G):
        """
        The inverse of G.

        Returns
        -------
        Ginv : Surreal
            The inverse of G.
        """
        return Surreal(1/G._n)



"""
Tests for ludolory.hypothesis.
"""

import pytest

from ludology import Game
from ludology.hypothesis import gamify


@pytest.mark.parametrize(['s', 'g'], [
    (([([], [([], [])]), ([], [])], [([], [])]), Game({0}, {0})),
    (Game({0}, {0}), Game({0}, {0})),
])
def test_gamify(s, g):
    """
    Test a contrived example.
    """
    assert gamify(s) == g



"""
Tests for ludology.sums.
"""

import pytest

from ludology import Game, canonicalize
from ludology.closet import zero, one, star, pm_one, up
from ludology.sums import (disjunctive, conjunctive, selective, diminished_disjunctive,
                           continued_conjunctive, shortened_selective, ordinal, side, sequential)


G = Game({0}, {pm_one})


@pytest.mark.parametrize(['G', 'H', 'J'], [
    (G, up, Game({1}, {Game({1}, {1})})),
    (G, zero, G),
    (zero, G, G),
])
def test_disjunctive(G, H, J):
    """
    Test that the disjunctive sum matches known results.
    """
    assert canonicalize(disjunctive(G, H)) == J


@pytest.mark.parametrize(['G', 'H', 'J'], [
    (G, up, up),
    (G, zero, zero),
    (zero, G, zero),
])
def test_conjunctive(G, H, J):
    """
    Test that the conjunctive sum matches known results.
    """
    assert canonicalize(conjunctive(G, H)) == J


@pytest.mark.parametrize(['G', 'H', 'J'], [
    (G, up, Game({1}, {Game({1, one + star}, {-1, -one + star}), Game({1}, {1, pm_one})})),
    (G, zero, G),
    (zero, G, G),
])
def test_selective(G, H, J):
    """
    Test that the selective sum matches known results.
    """
    assert canonicalize(selective(G, H)) == J


@pytest.mark.parametrize(['G', 'H', 'J'], [
    (G, up, up),
    (G, zero, zero),
    (zero, G, zero),
])
def test_diminished_disjunctive(G, H, J):
    """
    Test that the diminished disjunctive sum matches known results.
    """
    assert canonicalize(diminished_disjunctive(G, H)) == J


@pytest.mark.parametrize(['G', 'H', 'J'], [
    (G, up, one),
    (G, zero, G),
    (zero, G, G),
])
def test_continued_conjunctive(G, H, J):
    """
    Test that the continued conjunctive sum matches known results.
    """
    assert canonicalize(continued_conjunctive(G, H)) == J


@pytest.mark.parametrize(['G', 'H', 'J'], [
    (G, up, Game({0}, {star, Game({0, star}, {0, star})})),
    (G, zero, zero),
    (zero, G, zero),
])
def test_shortened_selective(G, H, J):
    """
    Test that the shortened selective sum matches known results.
    """
    assert canonicalize(shortened_selective(G, H)) == J


@pytest.mark.parametrize(['G', 'H', 'J'], [
    (G, up, Game({1}, {Game({1}, {1, pm_one}), pm_one})),
    (G, zero, G),
    (zero, G, G),
])
def test_ordinal(G, H, J):
    """
    Test that the ordinal sum matches known results.
    """
    assert canonicalize(ordinal(G, H)) == J


@pytest.mark.parametrize(['G', 'H', 'J'], [
    (G, up, Game({0}, {1/2})),
    (G, zero, G),
    (zero, G, G),
])
def test_side(G, H, J):
    """
    Test that the side sum matches known results.
    """
    assert canonicalize(side(G, H)) == J


@pytest.mark.parametrize(['G', 'H', 'J'], [
    (G, up, one),
    (G, zero, G),
    (zero, G, G),
])
def test_sequential(G, H, J):
    """
    Test that the sequential sum matches known results.
    """
    assert canonicalize(sequential(G, H)) == J



"""
Various utilities helpful in computing Game properties.
"""

from itertools import chain, combinations, count


__all__ = [
    'mex',
    'powerset',
]


def mex(s):
    """
    Compute the *m*inimum *ex*cluded element of the set `s`. The universe is assumed to be the
    nonnegative integers.

    Parameters
    ----------
    s : collection
        The set/collection to compute the mex of.

    Returns
    -------
    mex : int
        The smallest non-negative integer not included in `s`.
    """
    for i in count():  # pragma: no branch
        if i not in s:
            return i


def powerset(iterable):
    """
    Compute the powerset of the elements represented in `iterable`.

    Parameters
    ----------
    iterable : iterable
        The set of elements to compute the powerset of.

    Returns
    -------
    ps : iterable
        An iterator over the powerset of `iterable`.
    """
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))



"""
Tests for ludology.game.
"""

import pytest
from hypothesis import given
from hypothesis.strategies import integers

from ludology import Game
from ludology.closet import zero, half, one, star, star2, up, pm_one
from ludology.game import Outcome
from ludology.tools import canonicalize


@pytest.mark.parametrize(['g1', 'g2'], [
    (one, zero),
    (one, half),
    (one, -one),
    (zero, -one),
    (zero, -half),
    (up, zero),
    (half, up),
    (up + up, star),
    (zero, zero),
    (up, up),
])
def test_ge(g1, g2):
    """
    Test the ordering of several Games.
    """
    assert g1 >= g2


@pytest.mark.parametrize(['g1', 'g2'], [
    (one, zero),
    (one, half),
    (one, -one),
    (zero, -one),
    (zero, -half),
    (up, zero),
    (half, up),
    (up + up, star),
])
def test_gt(g1, g2):
    """
    Test the ordering of several Games.
    """
    assert g1 > g2


@pytest.mark.parametrize(['g1', 'g2'], [
    (zero, one),
    (half, one),
    (-one, one),
    (-one, zero),
    (-half, zero),
    (zero, up),
    (up, half),
    (star, up + up),
    (zero, zero),
    (up, up),
])
def test_le(g1, g2):
    """
    Test the ordering of several Games.
    """
    assert g1 <= g2


@pytest.mark.parametrize(['g1', 'g2'], [
    (zero, one),
    (half, one),
    (-one, one),
    (-one, zero),
    (-half, zero),
    (zero, up),
    (up, half),
    (star, up + up),
])
def test_lt(g1, g2):
    """
    Test the ordering of several Games.
    """
    assert g1 < g2


@pytest.mark.parametrize(['g1', 'g2'], [
    (star, zero),
    (star, up),
    (one, pm_one),
    (pm_one, -one),
    (pm_one, star),
    (pm_one, zero),
])
def test_fuzzy_1(g1, g2):
    """
    Test the ordering of several Games.
    """
    assert g1 | g2


def test_fuzzy_2():
    """
    Test the ordering of several Games.
    """
    assert star | 0
    assert 0 | star


def test_equiv_1():
    """
    Test the ordering of several Games.
    """
    g = Game({star})
    assert g == zero


def test_equiv_2():
    """
    Test the ordering of several Games.
    """
    g = Game(right={star})
    assert g == zero


@pytest.mark.parametrize('g', [
    zero,
    star,
    star2,
])
def test_impartial(g):
    """
    Test that these games are impartial.
    """
    assert g.is_impartial


@pytest.mark.parametrize('g', [
    one,
    half,
    up,
    up + star,
    pm_one,
])
def test_not_impartial(g):
    """
    Test that these games are not impartial.
    """
    assert not g.is_impartial


@pytest.mark.parametrize('g', [
    star,
    up,
    -up,
    Game({0}, {Game({0}, {-1})}),
])
def test_infinitesimal(g):
    """
    Test that these games are infinitesimal.
    """
    assert g.is_infinitesimal


@pytest.mark.parametrize('g', [
    half,
    one,
    zero,
    pm_one,
])
def test_not_infinitesimal(g):
    """
    Test that these games are not infinitesimal.
    """
    assert not g.is_infinitesimal


@pytest.mark.parametrize('g', [
    pm_one,
    Game({2}, {1}),
])
def test_switch(g):
    """
    Test that these games are switches.
    """
    assert g.is_switch


@pytest.mark.parametrize('g', [
    zero,
    one,
    up,
    star,
])
def test_not_switch(g):
    """
    Test that these games are not switches.
    """
    assert not g.is_switch


@pytest.mark.parametrize('g', [
    zero,
    star,
    up,
])
def test_dicotic(g):
    """
    Test that these games are dicotic.
    """
    assert g.is_dicotic


@pytest.mark.parametrize('g', [
    one,
    Game({0}, {Game({0}, {-1})}),
])
def test_not_dicotic(g):
    """
    Test that these games are not dicotic.
    """
    assert not g.is_dicotic


@pytest.mark.parametrize('g', [
    star,
    up,
    -up,
    one + star,
    -3 + up,
])
def test_numberish(g):
    """
    Test that these games are numberish.
    """
    assert g.is_numberish


@pytest.mark.parametrize('g', [
    pm_one,
    Game({5/2, Game({4}, {2})}, {Game({-1}, {-2}), Game({0}, {-4})}),
])
def test_not_numberish(g):
    """
    Test that these games are not numberish.
    """
    assert not g.is_numberish


@pytest.mark.parametrize(['g', 'bday'], [
    (zero, 0),
    (one, 1),
    (-one, 1),
    (star, 1),
    (up, 2),
    (canonicalize(up + star), 2),
    (half, 2),
])
def test_birthday(g, bday):
    """
    Test some birthdays.
    """
    assert g.birthday == bday


@pytest.mark.parametrize(['g', 'sp'], [
    (up, [zero, star, zero]),
])
def test_subpositions(g, sp):
    """
    """
    assert list(g.subpositions()) == sp


@pytest.mark.parametrize(['a', 'b', 'c'], [
    (zero, zero, zero),
    (star, star, zero),
    (one, -one, zero),
    (one, zero, one),
    (zero, one, one),
    (one, 0, one),
    (0, one, one),
    (one, star, Game({one}, {one})),
    (star, one, Game({one}, {one})),
])
def test_add(a, b, c):
    """
    Test Game addition.
    """
    assert a + b == c


@pytest.mark.parametrize(['a', 'b', 'c'], [
    (zero, zero, zero),
    (star, star, star),
    (one, -one, -1),
    (one + 1, 1, 2),
    (2, one, one + one),
])
def test_mul(a, b, c):
    """
    Test Game multiplication.
    """
    assert a * b == c


@pytest.mark.parametrize(['g', 'o'], [
    (zero, Outcome.PREVIOUS),
    (star, Outcome.NEXT),
    (one, Outcome.LEFT),
    (up, Outcome.LEFT),
    (-up, Outcome.RIGHT),
])
def test_outcome(g, o):
    """
    Test the outcome class of several games.
    """
    assert g.outcome == o


@pytest.mark.parametrize(['g', 'v'], [
    (star, '*'),
    (pm_one, '±1'),
    (Game({Game(3)}, {Game(1)}), '2±1'),
    (up, '↑'),
    (2 * up, '2·↑'),
    (up + star, '↑*'),
    (up + up + star, '2·↑*'),
    (-up, '↓'),
    (2 * -up, '2·↓'),
    (-up + star, '↓*'),
    (-up + -up + star, '2·↓*'),
    (Game({one}, {1}), '1*'),
    (Game({0}, {Game({0}, {-2})}), '⧾_2'),
    (Game({Game({2}, {0})}, {0}), '⧿_2'),
    (star2, '*2'),
    (Game({5/2, Game({4}, {2})}, {Game({-1}, {-2}), Game({0}, {-4})}), '{3±1,5/2|-2±2,-3/2±1/2}'),
    (Game({0}, {star2}), '{0|*2}'),
    (Game({star2}, {0}), '{*2|0}'),
])
def test_value(g, v):
    """
    Test that several games evaluate to the correct short-hand value.
    """
    assert g.value == v


@given(integers(min_value=-7, max_value=7), integers(min_value=0, max_value=3))
def test_value_dyadic_rational(m, j):
    """
    Test the construction of dyadic rationals.
    """
    f = m/2**j
    n, d = f.as_integer_ratio()
    g = Game(f)
    assert g.value == (f"{n}/{d}" if d > 1 else f"{n}")


def test_game_fail():
    """
    Test that Game punts of constructing "deep" games.
    """
    with pytest.raises(ValueError):
        Game(1/3)



"""
Tests for ludology.utils.
"""

import pytest

from ludology.utils import mex, powerset


@pytest.mark.parametrize(['s', 'v'], [
    ({1, 2, 3}, 0),
    ({0, 1, 3}, 2),
    ({0, 1, 2}, 3),
])
def test_mex(s, v):
    """
    Test that several mex's are correct.
    """
    assert mex(s) == v


@pytest.mark.parametrize('n', range(6))
def test_powerset(n):
    """
    Test that powersets have the correct size.
    """
    assert len(list(powerset(range(n)))) == 2**n



"""
A collection of common games.
"""

from .game import Game


__all__ = [
    'zero',
    'quarter',
    'half',
    'one',
    'two',
    'star',
    'star2',
    'up',
    'tiny',
    'pm_one',
    'g1',
    'g2',
    'g3',
    'g4',
]


# numbers
zero = Game()
quarter = Game(1/4)
half = Game(1/2)
one = Game(1)
two = Game(2)

# nimbers
star = Game({0}, {0})
star2 = Game({0, star}, {0, star})

# other infinitesimals
up = Game({0}, {star})
upstar = Game({0}, {0, star})
tiny = Game({0}, {Game({0}, {-1})})

# switch
pm_one = Game({1}, {-1})

# more complex examples
g1 = Game({5/2, Game({4}, {2})}, {Game({0}, {-4}), Game({-1}, {-2})})
g2 = Game({Game({4}, {2})}, {Game({Game({0}, {-2})}, {Game({-4}, {-6})})})
g3 = Game({Game({3}, {one + star})}, {Game({-one + star}, {-two + star}), Game({0}, {-3})})
g4 = Game({5}, {Game({1}, {-9})})



"""
Tests for ludology.surreal.
"""

import pytest

from ludology import Game, Surreal
from ludology.closet import zero, one, star


@pytest.mark.parametrize('n', [
    Surreal(0),
    Surreal(1/2),
    Surreal(0.125),
    Surreal(-1),
    Surreal(2),
])
def test_is_number(n):
    """
    Test that all Surreals are numbers.
    """
    assert n.is_number


@pytest.mark.parametrize(['n', 'v'], [
    (Surreal(0), True),
    (Surreal(1/2), False),
    (Surreal(0.125), False),
    (Surreal(-1), False),
    (Surreal(2), False),
])
def test_is_impartial(n, v):
    """
    Test that all Surreals (other than 0) are not impartial.
    """
    assert n.is_impartial == v


@pytest.mark.parametrize(['n', 'v'], [
    (Surreal(0), True),
    (Surreal(1/2), False),
    (Surreal(0.125), False),
    (Surreal(-1), False),
    (Surreal(2), False),
])
def test_is_dicotic(n, v):
    """
    Test that all Surreals (other than 0) are not dicotic.
    """
    assert n.is_dicotic == v


@pytest.mark.parametrize('n', [
    Surreal(0),
    Surreal(1/2),
    Surreal(0.125),
    Surreal(-1),
    Surreal(2),
])
def test_is_infinitesimal(n):
    """
    Test that all Surreals are not infinitesimal.
    """
    assert not n.is_infinitesimal


@pytest.mark.parametrize('n', [
    Surreal(0),
    Surreal(1/2),
    Surreal(0.125),
    Surreal(-1),
    Surreal(2),
])
def test_is_numberish(n):
    """
    Test that all Surreals are numberish.
    """
    assert n.is_numberish


@pytest.mark.parametrize('n', [
    Surreal(0),
    Surreal(1/2),
    Surreal(0.125),
    Surreal(-1),
    Surreal(2),
])
def test_is_switch(n):
    """
    Test that no Surreals are switches.
    """
    assert not n.is_switch


@pytest.mark.parametrize(['a', 'b'], [
    (0, 0),
    (1, 0),
    (-1, 0),
    (-1, 1),
    (1/2, 1/2),
    (1/2, 1),
])
def test_order(a, b):
    """
    Test the addition of two Surreals.
    """
    assert (Surreal(a) >= Surreal(b)) == (a >= b)


@pytest.mark.parametrize(['a', 'b'], [
    (Surreal(0), zero),
    (Surreal(1/2), zero),
    (Surreal(3/4), star),
])
def test_order_game(a, b):
    """
    Test the addition of two Surreals.
    """
    assert a >= b


@pytest.mark.parametrize(['a', 'b'], [
    (0, 0),
    (1, 0),
    (-1, 0),
    (-1, 1),
    (1/2, 1/2),
    (1/2, 1),
])
def test_add(a, b):
    """
    Test the addition of two Surreals.
    """
    assert Surreal(a) + Surreal(b) == Surreal(a + b)


@pytest.mark.parametrize(['a', 'b', 'c'], [
    (Surreal(1), star, one + star),
    (Surreal(2), 3 * one, Surreal(5)),
])
def test_add_game(a, b, c):
    """
    Test the addition of a Surreal with a Game.
    """
    assert a + b == c


@pytest.mark.parametrize(['a', 'b'], [
    (0, 0),
    (1, 0),
    (-1, 0),
    (-1, 1),
    (1/2, 1/2),
    (1/2, 1),
])
def test_add_game_2(a, b):
    """
    Test the addition of two Surreals.
    """
    assert Surreal(a) + Surreal(b) == Game(a + b)


@pytest.mark.parametrize(['a', 'b'], [
    (0, 0),
    (1, 0),
    (-1, 0),
    (-1, 1),
    (1/2, 1/2),
    (1/2, 1),
])
def test_mul(a, b):
    """
    Test the multiplication of two Surreals.
    """
    assert Surreal(a) * Surreal(b) == Surreal(a * b)


@pytest.mark.parametrize(['a', 'b'], [
    (0, 0),
    (1, 0),
    (-1, 0),
    (-1, 1),
    (1/2, 1/2),
    (1/2, 1),
])
def test_mul_game(a, b):
    """
    Test the multiplication of two Surreals.
    """
    assert Surreal(a) * Game(b) == Surreal(a * b)


@pytest.mark.parametrize(['a', 'b'], [
    (0, 0),
    (1, 0),
    (-1, 0),
    (-1, 1),
    (1/2, 1/2),
    (1/2, 1),
])
def test_mul_game_2(a, b):
    """
    Test the multiplication of two Surreals.
    """
    assert Surreal(a) * Surreal(b) == Game(a * b)


@pytest.mark.parametrize(['a', 'b'], [
    (1, 1),
    (1, 1/2),
    (2, 1),
])
def test_div(a, b):
    """
    Test the division of two Surreals.
    """
    assert Surreal(a) / Surreal(b) == Surreal(a / b)


@pytest.mark.xfail
@pytest.mark.parametrize(['a', 'b'], [
    (1, 1),
    (1, 1/2),
    (2, 1),
])
def test_div_game(a, b):
    """
    Test the division of a Surreal by a Game.
    """
    assert Surreal(a) / Game(b) == Game(a / b)


@pytest.mark.parametrize('n', [
    1,
    2,
    8,
    1/2,
    1/4,
])
def test_invert(n):
    """
    Test the inversion of a Surreal.
    """
    assert Surreal(n)._invert() == Surreal(1/n)


def test_surreal_fail():
    """
    Test that Surreal doesn't try to construct things like 1/3.
    """
    with pytest.raises(ValueError):
        Surreal(1/3)



"""
Functions related to the calculation of thermographic properties.
"""

from collections import namedtuple
from copy import copy
from fractions import Fraction
from functools import lru_cache

import numpy as np
from scipy.optimize import brentq, minimize

from .game import Game
from .tools import canonicalize


__al__ = [
    'is_cold',
    'is_tepid',
    'is_hot',
    'mean',
    'temperature',
    'cool',
    'heat',
    'overheat',
    'thermal_dissociation',
    'thermograph',
]


###############################################################################
# hot, cold, or tepid? [Siegel pg. 112]


def is_cold(G):
    """
    A Game is cold if it is a number.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    cold : bool
        Whether the Game is cold or not.
    """
    return G.is_number


def is_tepid(G):
    """
    A Game is tepid if it is numberish, but not a number; that is, it differs from a number by an
    infinitesimal amount.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    tepid : bool
        Whether the Game is tepid or not.
    """
    return G.is_numberish and not G.is_number


def is_hot(G):
    """
    A Game is hot if it is not numberish; that is, it's left and right stops do not coincide.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    hot : bool
        Whether the Game is hot or not.
    """
    return not G.is_numberish


###############################################################################
# mean and temperature


@lru_cache(maxsize=None)
def mean(G):
    """
    Compute the mean value of a Game. The mean value is defined as the value to which a Game cools
    to. It is, by definition, a surreal number.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    m : float
        The mean value of G.
    """
    value = G.value
    if G.is_number:
        return float(Fraction(value))
    elif G.is_switch:
        m, _ = value.split('±')
        if m:
            m = float(Fraction(m))
        else:
            m = 0.0
        return m
    else:
        return float(_left_scaffold(G)(temperature(G)))


@lru_cache(maxsize=None)
def temperature(G):
    """
    Compute the temperature of a game. It is the amount by which a Game must be cooled in order to
    become a number.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    temp : float
        The temperature of G.
    """
    if G.is_number:
        return -1/Fraction(G.value).denominator
    elif G.is_switch:
        _, temp = G.value.split('±')
        temp = float(Fraction(temp))
        return temp
    else:
        fl = _left_scaffold(G)
        fr = _right_scaffold(G)

        def f1(t):
            return fl(t) - fr(t)

        def f2(t):
            return abs(fl(t) - fr(t)) + t/2

        upper = 1
        while f1(upper) > 0:
            upper *= 2

        root = float(brentq(f1, 0, upper + 1))

        return round(float(minimize(f2, root).x), 4)


###############################################################################
# heating and cooling of games


@lru_cache(maxsize=None)
def cool(G, t):
    """
    Cool a game down. To cool a game, one taxes each player for the right to play. This results in
    both players having less incentive to play. At some point both the left stop and the right stop
    of the game coincide, at which point the game has become tepid (infinitesimally close to a
    number) and cooling by any more will make the game cold (a number).

    Parameters
    ----------
    G : Game
        The Game of interest.
    t : Game, float
        The amount to cool G by.

    Returns
    -------
    CG : Game
        The Game resulting from cooling G by t.
    """
    if t <= temperature(G):
        t = Game(t)
        lefts = {cool(G_L, t) - t for G_L in G._left}
        rights = {cool(G_R, t) + t for G_R in G._right}
        return Game(lefts, rights)
    else:
        return Game(mean(G))


@lru_cache(maxsize=None)
def heat(G, t):
    """
    Heat up a game. In some sense, heating is the opposite of cooling.

    Parameters
    ----------
    G : Game
        The Game of interest.
    t : Game, float
        The amount to heat G by.

    Returns
    -------
    HG : Game
        The Game resulting from heating G by t.
    """
    if G.is_number:
        return G
    else:
        if not isinstance(t, Game):
            t = Game(t)
        lefts = {heat(G_L, t) + t for G_L in G._left}
        rights = {heat(G_R, t) - t for G_R in G._right}
        return Game(lefts, rights)


@lru_cache(maxsize=None)
def overheat(G, t):
    """
    Overheat a game. This is similar to heating, except applies to numbers also.

    Parameters
    ----------
    G : Game
        The Game of interest.
    t : Game, float
        The amount to overheat G by.

    Returns
    -------
    HG : Game
        The Game resulting from overheating G by t.
    """
    if not isinstance(t, Game):
        t = Game(t)
    lefts = {overheat(G_L, t) + t for G_L in G._left}
    rights = {overheat(G_R, t) - t for G_R in G._right}
    return Game(lefts, rights)


###############################################################################
# thermal dissociation


Particle = namedtuple('Particle', ['particle', 'critical_temperature'])


def thermal_dissociation(G):
    """
    The thermal dissociation of a Game G is its mean value, plus a series of
    heated infinitesimals.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    td : tuple
        A tuple consisting of the mean value, followed by Particles.
    """
    g = copy(G)

    m = mean(g)

    particles = []

    while not g.is_infinitesimal and g != 0:
        t = temperature(g)
        p = canonicalize(cool(g, t) - mean(g))
        particles.append(Particle(p, t))
        g = canonicalize(g - heat(p, t))

    return tuple([m] + particles)


###############################################################################
# scaffolds, mostly helpers for thermograph plotting


def cooled_left_stop(G):  # pragma: no cover
    """
    The left stop of `G` cooled down. Useful for plotting thermographs.

    Parameters
    ----------
    G : Game
        The game of interest.

    Returns
    -------
    cls : func
        A function which returns the left stop of `G` cooled by `t`.
    """
    m = mean(G)
    crss = [cooled_right_stop(G_L) for G_L in G._left]

    @np.vectorize
    def inner(t):
        """
        Compute the cooled left stop.

        Parameters
        ----------
        t : float
            The amount to cool by.

        Returns
        -------
        cls : float
            The left stop of the cooled Game.
        """
        return max([crs(t) - t for crs in crss] + [m])

    return inner


def cooled_right_stop(G):  # pragma: no cover
    """
    The right stop of `G` cooled down. Useful for plotting thermographs.

    Parameters
    ----------
    G : Game
        The game of interest.

    Returns
    -------
    crs : func
        A function which returns the right stop of `G` cooled by `t`.
    """
    m = mean(G)
    clss = [cooled_left_stop(G_R) for G_R in G._right]

    @np.vectorize
    def inner(t):
        """
        Compute the cooled right stop.

        Parameters
        ----------
        t : float
            The amount to cool by.

        Returns
        -------
        cls : float
            The right stop of the cooled Game.
        """
        return min([cls(t) + t for cls in clss] + [m])

    return inner


def _left_scaffold(G):
    """
    Compute the left scaffold of the thermograph of G.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    ls : func
        The left scaffold as a function of temperature.
    """
    crss = [cooled_right_stop(G_L) for G_L in G._left]

    @np.vectorize
    def inner(t):
        """
        Compute the left scaffold of G as a function t.

        Parameters
        ----------
        t : float
            The temperature.

        Returns
        -------
        ls : float
            The left scaffold of G at temperature t.
        """
        return np.max([crs(t) - t for crs in crss])

    return inner


def _right_scaffold(G):
    """
    Compute the right scaffold of the thermograph of G.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    rs : func
        The right scaffold as a function of temperature.
    """
    clss = [cooled_left_stop(G_R) for G_R in G._right]

    @np.vectorize
    def inner(t):
        """
        Compute the right scaffold of G as a function t.

        Parameters
        ----------
        t : float
            The temperature.

        Returns
        -------
        rs : float
            The right scaffold of G at temperature t.
        """
        return np.min([cls(t) + t for cls in clss])

    return inner


@lru_cache(maxsize=None)
def number_height(G):  # pragma: no cover
    """
    Return the height of the game tree truncated at numbers.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    height : int
        The hight of the number-truncated game tree.
    """
    if G.is_number:
        return 1
    else:
        return max(number_height(g) for g in G._left | G._right) + 1


###############################################################################
# plotting of thermographs


def _thermograph(G, ax, t_min, t_max, lw):  # pragma: no cover
    """
    Plot a single thermograph.

    Parameters
    ----------
    G : Game
        The Game whose thermograph is to be plotted.
    ax : plt.axis
        The matplotlib axis to plot on.
    t_min : float
        The minimum temperature to plot.
    t_max : float
        The maximum temperature to plot.
    lw : float
        The line width to plot with.
    """
    temp = temperature(G)
    ts = np.linspace(t_min, temp, 101)
    ls, rs = cooled_left_stop(G), cooled_right_stop(G)
    mast_height = t_max - temp

    line, = ax.plot(ls(ts), ts, lw=lw, label=G.value)
    color = line.get_color()
    ax.plot(rs(ts), ts, lw=lw, c=color)

    ax.arrow(mean(G), temp, 0, mast_height, lw=lw, color=color, length_includes_head=True)


def thermograph(G, with_options=True, ax=None):  # pragma: no cover
    """
    Plot a thermograph of G.

    Parameters
    ----------
    G : Game
        The Game to plot the thermograph of.
    with_options : bool
        If True, plot the thermographs of the options of G as well. Defaults to True.
    ax : plt.axis, None
        The axis to plot on. If not provided, construct one.

    Returns
    -------
    ax : plt.axis
        The axis with the thermograph plotted.
    """
    if ax is None:
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=(10, 5))
        ax = fig.gca()

    t_min = -1.0 if G.is_number else 0.0
    t_max = max([0.25 + temperature(G), 1.0])

    if with_options:
        lw = 2 * number_height(G)
    else:
        lw = 2

    _thermograph(G, ax, t_min, t_max, lw)

    if with_options:
        options = G._left | G._right
        i = 2
        while any(not g.is_number for g in options):
            new_options = set()
            for g in options:
                _thermograph(g, ax, t_min, t_max, lw - i)
                g_opts = g._left | g._right
                if any(not _.is_number for _ in g_opts):
                    new_options |= g_opts
            i += 2
            options = new_options

    ax.invert_xaxis()
    ax.set_ylim(t_min, t_max + 0.1 * (t_max - t_min))

    ax.grid(True)

    ax.legend(loc='best')

    return ax



"""
A variety of ways of summing games.
"""

from functools import lru_cache, wraps

from .game import Game
from .tools import canonicalize

__all__ = [
    'disjunctive',
    'conjunctive',
    'selective',
    'diminished_disjunctive',
    'continued_conjunctive',
    'shortened_selective',
    'ordinal',
    'side',
    'sequential',
]


def canon(f):
    """
    Add a `canon` option to `f` which toggles canonicalizes the return value of
    `f`.

    Parameters
    ----------
    f : func
        The function to wrap.

    Returns
    -------
    wrapped : func
        The wrapped form of `f`.
    """
    @wraps(f)
    def wrapped(G, H, canon=True):
        game = f(G, H)
        if canon:
            game = canonicalize(game)
        return game

    return wrapped


@lru_cache(maxsize=None)
@canon
def disjunctive(G, H):
    """
    Move in exactly one component.

    Parameters
    ----------
    G : Game
        The first Game.
    H : Game
        The second Game.

    Returns
    -------
    sum : Game
        The disjunctive sum of G and H.
    """
    left_1 = {disjunctive(G_L, H) for G_L in G._left}
    left_2 = {disjunctive(G, H_L) for H_L in H._left}
    right_1 = {disjunctive(G_R, H) for G_R in G._right}
    right_2 = {disjunctive(G, H_R) for H_R in H._right}
    return Game(left_1 | left_2, right_1 | right_2)


@lru_cache(maxsize=None)
@canon
def conjunctive(G, H):
    """
    Move in all components. Play ends when any one of them terminates.

    Parameters
    ----------
    G : Game
        The first Game.
    H : Game
        The second Game.

    Returns
    -------
    sum : Game
        The conjunctive sum of G and H.
    """
    left = {conjunctive(G_L, H_L) for G_L in G._left for H_L in H._left}
    right = {conjunctive(G_R, H_R) for G_R in G._right for H_R in H._right}
    return Game(left, right)


@lru_cache(maxsize=None)
@canon
def selective(G, H):
    """
    Move in any number of components, but at least one.

    Parameters
    ----------
    G : Game
        The first Game.
    H : Game
        The second Game.

    Returns
    -------
    sum : Game
        The selective sum of G and H.
    """
    left_1 = {selective(G_L, H) for G_L in G._left}
    left_2 = {selective(G, H_L) for H_L in H._left}
    left_3 = {selective(G_L, H_L) for G_L in G._left for H_L in H._left}
    right_1 = {selective(G_R, H) for G_R in G._right}
    right_2 = {selective(G, H_R) for H_R in H._right}
    right_3 = {selective(G_R, H_R) for G_R in G._right for H_R in H._right}
    return Game(left_1 | left_2 | left_3, right_1 | right_2 | right_3)


@lru_cache(maxsize=None)
@canon
def diminished_disjunctive(G, H):
    """
    Move in exactly one component. Play ends immediately when any one of them terminates.

    Parameters
    ----------
    G : Game
        The first Game.
    H : Game
        The second Game.

    Returns
    -------
    sum : Game
        The diminished disjunctive sum of G and H.
    """
    if G == 0 or H == 0:
        return Game(0)
    else:
        left_1 = {diminished_disjunctive(G_L, H) for G_L in G._left}
        left_2 = {diminished_disjunctive(G, H_L) for H_L in H._left}
        right_1 = {diminished_disjunctive(G_R, H) for G_R in G._right}
        right_2 = {diminished_disjunctive(G, H_R) for H_R in H._right}
        return Game(left_1 | left_2, right_1 | right_2)


@lru_cache(maxsize=None)
@canon
def continued_conjunctive(G, H):
    """
    Move in all nonterminal components. Play ends only after all components terminate.

    Parameters
    ----------
    G : Game
        The first Game.
    H : Game
        The second Game.

    Returns
    -------
    sum : Game
        The continued conjunctive sum of G and H.
    """
    if G == 0 or H == 0:
        return disjunctive(G, H)
    else:
        left = {continued_conjunctive(G_L, H_L) for G_L in G._left for H_L in H._left}
        right = {continued_conjunctive(G_R, H_R) for G_R in G._right for H_R in H._right}
        return Game(left, right)


@lru_cache(maxsize=None)
@canon
def shortened_selective(G, H):
    """
    Move in any number of components. Play ends immediately when any one of them terminates.

    Parameters
    ----------
    G : Game
        The first Game.
    H : Game
        The second Game.

    Returns
    -------
    sum : Game
        The shortened selective sum of G and H.
    """
    if G == 0 or H == 0:
        return Game(0)
    else:
        left_1 = {shortened_selective(G_L, H) for G_L in G._left}
        left_2 = {shortened_selective(G, H_L) for H_L in H._left}
        left_3 = {shortened_selective(G_L, H_L) for G_L in G._left for H_L in H._left}
        right_1 = {shortened_selective(G_R, H) for G_R in G._right}
        right_2 = {shortened_selective(G, H_R) for H_R in H._right}
        right_3 = {shortened_selective(G_R, H_R) for G_R in G._right for H_R in H._right}
        return Game(left_1 | left_2 | left_3, right_1 | right_2 | right_3)


@lru_cache(maxsize=None)
@canon
def ordinal(G, H):
    """
    Move in G or H; any move on G annihilates H.

    Parameters
    ----------
    G : Game
        The first Game.
    H : Game
        The second Game.

    Returns
    -------
    sum : Game
        The ordinal sum of G and H.
    """
    left_1 = {G_L for G_L in G._left}
    left_2 = {ordinal(G, H_L) for H_L in H._left}
    right_1 = {G_R for G_R in G._right}
    right_2 = {ordinal(G, H_R) for H_R in H._right}
    return Game(left_1 | left_2, right_1 | right_2)


@lru_cache(maxsize=None)
@canon
def side(G, H):
    """
    Move in G or H; Left's moves on H annihilate G, and Right's moves on G annihilate H.

    Parameters
    ----------
    G : Game
        The first Game.
    H : Game
        The second Game.

    Returns
    -------
    sum : Game
        The side sum of G and H.
    """
    left_1 = {side(G_L, H) for G_L in G._left}
    left_2 = {H_L for H_L in H._left}
    right_1 = {G_R for G_R in G._right}
    right_2 = {side(G, H_R) for H_R in H._right}
    return Game(left_1 | left_2, right_1 | right_2)


@lru_cache(maxsize=None)
@canon
def sequential(G, H):
    """
    Move in G unless G has terminated; in that case move in H.

    Parameters
    ----------
    G : Game
        The first Game.
    H : Game
        The second Game.

    Returns
    -------
    sum : Game
        The sequential sum of G and H.
    """
    if G == 0:
        return H
    else:
        left = {sequential(G_L, H) for G_L in G._left}
        right = {sequential(G_R, H) for G_R in G._right}
        return Game(left, right)



"""
Tests for ludology.lattice.
"""

import pytest

from ludology import Game
from ludology.closet import zero, one, star, up, upstar
from ludology.lattice import all_games, companion, is_lonely


def test_all_games_gen_1():
    """
    Test that the games generated by day 1 are correct.
    """
    gen1a = {zero, one, -one, star}
    gen1b = all_games(1)
    assert gen1a == gen1b


def test_all_games_gen_2():
    """
    Test that the correct number of games are crated by day 2.
    """
    gen2 = all_games(2)
    assert len(gen2) == 22


@pytest.mark.parametrize(['g1', 'g2'], [
    (zero, star),
    (one, one),
    (up, -upstar),
    (Game({-1}, {-3}), Game({-1}, {-3})),
])
def test_companion(g1, g2):
    """
    Test companionship.
    """
    assert companion(g1) == g2


@pytest.mark.parametrize('g', [
    one,
    Game({-1}, {-3}),
])
def test_is_lonely(g):
    """
    Test that certain Games are lonely.
    """
    assert is_lonely(g)



"""
This module implements functions related to the structure of the partial order of games.
"""

from itertools import combinations, product

from .game import Game
from .tools import canonicalize
from .utils import powerset


__all__ = [
    'all_games',
    'companion',
    'is_lonely',
]

def is_antichain(stuff):
    """
    Whether the elements of `stuff` form an antichain or not. An antichain is a set where no two
    elements are comparable.

    Parameters
    ----------
    stuff : collection
        The collection of elements to test.

    Returns
    -------
    antichain : bool
        Whether `stuff` forms an antichain or not.
    """
    return all(g | h for g, h in combinations(stuff, 2))


def all_games(generation):
    """
    Construct all games born by day `generation`.

    Parameters
    ----------
    generation : int >= 0
        The generation of games to construct.

    Returns
    -------
    games : set
        The set of all games born by day `n`.
    """
    if generation == 0:
        return {Game()}
    else:
        priors = all_games(generation-1)
        antichains = {games for games in powerset(priors) if is_antichain(games)}
        news = {canonicalize(Game(a, b)) for a, b in product(antichains, repeat=2)}
        return priors | news


def companion(G):
    """
    The companion of a Game G is the dual of G under the unique order-preserving automorphism of
    Games born by day n.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    c : Game
        G's companion.
    """
    cleft = {companion(G_L) for G_L in G._left}
    cright = {companion(G_R) for G_R in G._right}
    if G == 0:
        c = Game({0} | cleft, {0} | cright)
    elif G > 0:
        c = Game({0} | cleft, cright)
    elif G < 0:
        c = Game(cleft, {0} | cright)
    else:
        c = Game(cleft, cright)
    return canonicalize(c)


def is_lonely(G):
    """
    A Game is lonely if it is its own companion.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    lonely : bool
        Whether G is lonely or not.
    """
    return G == companion(G)



"""
Nimbers are the values of impartial games.
"""

from functools import lru_cache
from itertools import count
from math import inf

from .game import Game
from .utils import mex


__all__ = [
    'Nimber',
    'FarStar',
]


class Nimber(Game):
    """
    A Nimber is the value of an impartial game.
    """
    def __init__(G, n):
        """
        Construct the `n`th Nimber.

        Parameters
        ----------
        n : int >= 0
            The value of Nimber to construct.
        """
        if n == 0:
            options = set()
        elif isinstance(n, int):
            options = {Nimber(i) for i in range(n)}
        else:
            msg = "Nimbers must have an integer value."
            raise ValueError(msg)

        G._left = G._right = options
        G._n = n

    @property
    def value(G):
        """
        Return the value of the Nimber.

        Returns
        -------
        v : str
            The value of `G`.
        """
        if G._n == 0:
            return '0'
        elif G._n == 1:
            return '*'
        else:
            return f"*{G._n}"

    @property
    def is_number(G):
        """
        Whether the Nimber is a number or not. Only the Nimber 0 is a number.

        Returns
        -------
        number : bool
            Whether the Nimber is a number or not.
        """
        return G._n == 0

    @property
    def is_numberish(G):
        """
        A Game is numberish if it is infinitesimally close to a number. All Nimbers are numberish.

        Returns
        -------
        numberish : bool
            Whether G is numberish or not (it is).
        """
        return True

    @property
    def is_impartial(G):
        """
        Whether the Nimber is impartial or not. All Nimbers are impartial.

        Returns
        -------
        impartial : bool
            Whether the Nimber is impartial or not.
        """
        return True

    @property
    def is_infinitesimal(G):
        """
        A game is infinitesimal if it is non-zero and smaller than any positive number and greater
        than any negative number. Equivalently, it's left and right stops are both zero. Note, this
        does not imply that an infinitesimal can not be positive (> 0) or negative (< 0).

        Returns
        -------
        infinitesimal : bool
            Whether the Nimber is infinitesimal or not (it is, unless it is zero).
        """
        return G._n != 0

    @property
    def is_dicotic(G):
        """
        A dicotic, or all-small, Game is one where either both or neither player have options
        at every subposition. All Nimbers are dicotic by definition of an impartial game.

        Returns
        -------
        dicotic : bool
            Whether the Game is dicotic or not (it is).
        """
        return True

    @property
    def birthday(G):
        """
        The Nimber's birthday.

        Returns
        -------
        bday : int >= 0
            The Nimber's birthday.
        """
        return G._n

    def __neg__(G):
        """
        The negation of the Nimber `G`. A Nimber is its own inverse.

        Returns
        -------
        G : Nimber
            The Nimber's inverse.
        """
        return G

    @lru_cache(maxsize=None)
    def __add__(G, H):
        """
        The sum of two Nimbers.

        Parameters
        ----------
        H : Nimber, Game
            The Nimber or Game to add by.

        Returns
        -------
        sum : Nimber, Game
            The sum of G and H.
        """
        if isinstance(H, Nimber):
            return Nimber(G._n ^ H._n)
        elif isinstance(H, Game):
            return super().__add__(H)
        else:
            return NotImplemented

    @lru_cache(maxsize=None)
    def __mul__(G, H):
        """
        The product of two Nimbers.

        Parameters
        ----------
        H : Nimber, Game
            The Nimber or Game to multiply by.

        Returns
        -------
        prod : Nimber, Game
            The product of G and H.
        """
        if isinstance(H, Nimber):
            return Nimber(_mul(G._n, H._n))
        elif isinstance(H, Game):
            return super().__mul__(H)
        else:
            return NotImplemented

    def __eq__(G, H):
        """
        The equality of two Nimbers.

        Parameters
        ----------
        H : Nimber, Game
            The Nimber or Game equate to `G`.

        Returns
        -------
        eq : Nimber, Game
            The equality of G and H.
        """
        if isinstance(H, Nimber):
            return G._n == H._n
        elif isinstance(H, Game):
            return super().__eq__(H)
        else:
            return NotImplemented

    def __hash__(G):
        """
        Define the hash of a Nimber as the hash of its value.

        Returns
        -------
        hash : str
            The hash of G.
        """
        return hash(G._n)


class FarStar(Nimber):
    """
    The far-star represents an arbitrarily large Nimber.
    """
    def __init__(G):
        """
        """
        G._n = inf

    @property
    def _left(G):
        """
        The "left set" of far-star.
        """
        return (Nimber(i) for i in count())  # pragma: no branch

    @property
    def _right(G):
        """
        The "right set" of far-star.
        """
        return (Nimber(i) for i in count())  # pragma: no branch

    @property
    def value(G):
        """
        A white star is used to represented far-star.
        """
        return "☆"

    def __add__(G, H):
        """
        The sum of far-star with any Nimber is far-star.

        Parameters
        ----------
        H : Nimber
            The Nimber to add to far-star.

        Returns
        -------
        fs : FarStar
            The far-star.
        """
        if isinstance(H, Nimber):
            return FarStar()
        else:
            return NotImplemented


def _mul(n1, n2):
    """
    The nimber product of `n1` and `n2`.

    Parameters
    ----------
    n1 : int
        The first nimber.
    n2 : int
        The second nimber.

    Returns
    -------
    prod : int
        The
    """
    return mex({_mul(a, n2) ^ _mul(n1, b) ^ _mul(a, b) for a in range(n1) for b in range(n2)})



"""
Tests for ludology.thermography.
"""

import pytest

from ludology import Game, Surreal
from ludology.closet import zero, quarter, half, one, pm_one, up, star
from ludology.thermography import (mean, temperature, cool, heat, overheat,
                                   is_cold, is_tepid, is_hot)


G_L1 = Game(5/2)
G_L2 = Game({Game(4)}, {Game(2)})
G_R1 = Game({Game(-1)}, {Game(-2)})
G_R2 = Game({Game()}, {Game(-4)})
g1 = Game({G_L1, G_L2}, {G_R1, G_R2})


@pytest.mark.parametrize(['g', 'm'], [
    (pm_one, 0.0),
    (one + pm_one, 1.0),
    (3 * one, 3.0),
    (g1, 0.5),
])
def test_mean(g, m):
    """
    Test that several mean values are correct.
    """
    assert mean(g) == m


@pytest.mark.parametrize(['g', 't'], [
    (pm_one, 1.0),
    (3 * one, -1.0),
    (g1, 2.5),
    (half, -1/2),
    (half + quarter, -1/4),
])
def test_temperature(g, t):
    """
    Test that several temperatures are correct.
    """
    assert temperature(g) == t


@pytest.mark.parametrize(['g', 't', 'v'], [
    (Game({2}, {-2}), 1.0, Game({1}, {-1})),
    (g1, 2.0, Game({1}, {Game({0}, {0})})),
])
def test_cooling(g, t, v):
    """
    Test that several cooled Games are correct.
    """
    assert cool(g, t) == v


@pytest.mark.parametrize(['g', 't', 'v'], [
    (pm_one, one, Game({2}, {-2})),
    (pm_one, 1, Game({2}, {-2})),
    (2 * one, 1.0, 2 * one),
])
def test_heating(g, t, v):
    """
    Test that several heated Games are correct.
    """
    assert heat(g, t) == v


@pytest.mark.parametrize(['g', 't', 'v'], [
    (pm_one, 1.0, Game({3}, {-3})),
    (star, 1.0, pm_one),
    (up, 1.0, Game({1}, {Game({0}, {-2})})),
])
def test_overheating(g, t, v):
    """
    Test that several overheated Games are correct.
    """
    assert overheat(g, t) == v


@pytest.mark.parametrize('g', [
    3 * one,
    Surreal(1/4),
    zero,
    -4 * one,
])
def test_is_cold(g):
    """
    Assert that Numbers are cold.
    """
    assert is_cold(g)


@pytest.mark.parametrize('g', [
    star,
    2 + star,
    5 * one + up,
])
def test_is_tepid(g):
    """
    Assert that Numbers are tepid.
    """
    assert is_tepid(g)


@pytest.mark.parametrize('g', [
    pm_one,
    g1,
])
def test_is_hot(g):
    """
    Assert that Numbers are hot.
    """
    assert is_hot(g)



"""
Various tools for computing properties of Games.
"""

from copy import copy
from functools import lru_cache


__all__ = [
    'canonicalize',
    'left_incentives',
    'right_incentives',
    'left_stop',
    'right_stop',
]


@lru_cache(maxsize=None)
def left_incentives(G):
    """
    Compute the left incentives of G:

        { G_L - G }

    Parameters
    ----------
    G : Game
        The game of interest.

    Returns
    -------
    incentives : set
        The left incentives of G.
    """
    return {canonicalize(G_L - G) for G_L in G._left}


@lru_cache(maxsize=None)
def right_incentives(G):
    """
    Compute the right incentives of G:

        { G - G_R }

    Parameters
    ----------
    G : Game
        The game of interest.

    Returns
    -------
    incentives : set
        The right incentives of G.
    """
    return {canonicalize(G - G_R) for G_R in G._right}


def stop_order(item):
    """
    if x > y:
        x_- > x_+ > y_- > y_+

    since L is lexicographically less than R, we
    can't rely on natural ordering.
    """
    g, s = item
    return (g, s == '-')


@lru_cache(maxsize=None)
def left_stop(G, adorn=True):
    """
    Compute the (adorned) left stop of G.

    Parameters
    ----------
    G : Game
        The Game of interest.
    adorn : bool
        Whether to adorn the stop or not. Defaults to True.

    Returns
    -------
    LS : Game
        The left stop of G.
    initiative : str
        '+' if it is left's turn, '-' if right's. Only returned if adorn is True.
    """
    if G.is_number:
        ls = (G, '+')
    else:
        ls = max((right_stop(G_L) for G_L in G._left), key=stop_order)

    if adorn:
        return ls
    else:
        return ls[0]


@lru_cache(maxsize=None)
def right_stop(G, adorn=True):
    """
    Compute the (adorned) right stop of G.

    Parameters
    ----------
    G : Game
        The Game of interest.
    adorn : bool
        Whether to adorn the stop or not. Defaults to True.

    Returns
    -------
    RS : Game
        The right stop of G.
    initiative : str
        '+' if it is left's turn, '-' if right's. Only returned if adorn is True.
    """
    if G.is_number:
        rs = (G, '-')
    else:
        rs = min((left_stop(G_R) for G_R in G._right), key=stop_order)

    if adorn:
        return rs
    else:
        return rs[0]


def remove_dominated(G):
    """
    Remove the dominated options of G.

    A left option is dominated if there exists another left option greater than
    it. A right option is dominated if there exists another right option less
    than it. In essence, it would always be a "bad idea" to move to a dominated
    option, because there exists a different option which was objectively and
    strictly superior.

    Parameters
    ----------
    G : Game
        The Game of interest.
    """
    G._left = {g for g in G._left if not any(g < G_L for G_L in G._left)}
    G._right = {g for g in G._right if not any(g > G_R for G_R in G._right)}


def replace_reversible(G):
    """
    Remove the reversable options of G.

    A left option is reversible if it has a right option which is less than G.
    In this case, that left option can be replaced with it's right option's left options.
    Essentially, this means that if left were to make a move to an option from which right
    has the ability to move to a position which is strictly better for her than G, it is a
    "no brainer" to do so, and so that left option might as well be replaced with the options
    available after right makes the obvious responce.

    Parameters
    ----------
    G : Game
        The Game of interest.
    """
    new_left_set = set()
    for G_L in G._left:
        for G_LR in G_L._right:
            if G_LR <= G:  # G_L is reversible through G_LR
                for G_LRL in G_LR._left:
                    new_left_set.add(G_LRL)
                break
        else:  # Not reversible
            new_left_set.add(G_L)
    G._left = new_left_set

    new_right_set = set()
    for G_R in G._right:
        for G_RL in G_R._left:
            if G_RL >= G:  # G_R is reversible through G_RL
                for G_RLR in G_RL._right:
                    new_right_set.add(G_RLR)
                break
        else:  # Not reversible
            new_right_set.add(G_R)
    G._right = new_right_set


def canonicalize(G):
    """
    Return the canonical form of the game G.

    Parameters
    ----------
    G : Game
        The Game of interest.

    Returns
    -------
    K : Game
        The Game G in canonical form.
    """
    G = copy(G)

    G._left = {canonicalize(G_L) for G_L in G._left}
    G._right = {canonicalize(G_R) for G_R in G._right}

    old_left, old_right = set(), set()
    while G._left != old_left or G._right != old_right:
        old_left, old_right = G._left, G._right
        remove_dominated(G)
        replace_reversible(G)

    return G



"""
A hypothesis strategy for sampling Games.

We use hypothesis.strategies.recursive to build pairs of options recursively, with some base
set of Games as leaves. By manual testing, using the second day games as leaves, and limiting left
and right options to having at most two games tends to sample things best. Having too many options
increases the likelihood of a number which would dominate any interesting infinitesimals. As such,
it would be pretty rare to sample tinies or the like.
"""

from hypothesis.strategies import composite, lists, recursive, sampled_from, tuples

from .game import Game
from .lattice import all_games
from .tools import canonicalize


__all__ = [
    'games',
]


def gamify(options):
    """
    Transform a nested pair of options into a Game in canonical form.

    Parameters
    ----------
    options : tuple(list, list)
        Nested tuples containing two lists of tuples containing two lists... with Games as leaves.

    Returns
    -------
    g : Game
        The Game represented by `options`.
    """
    try:
        left, right = options
        left = [gamify(_) for _ in left]
        right = [gamify(_) for _ in right]
        return canonicalize(Game(left, right))
    except TypeError:
        return options


@composite
def games(draw, base_day=2, max_options=2):  # pragma: no cover
    """
    A Hypothesis strategy for generating Games.

    Parameters
    ----------
    draw : func
        Required by hypothesis.strategies.composite.
    base_day : int >= 0
        The maximum birthday of games to use a leaves. Defaults to 2.
    max_options : int >= 0
        The maximum number of left, right options to allow for each subposition. Defaults to 2.
        Note that setting this number larger likely leads to the set being dominated by a number,
        reducing the variability in sampled games.

    Returns
    -------
    sample : Game
        A random Game.
    """
    options = draw(recursive(base=sampled_from(list(all_games(base_day))),
                             extend=lambda child: tuples(lists(child, max_size=max_options),
                                                         lists(child, max_size=max_options))))

    return gamify(options)



"""
The basic Game class.
"""

from collections import namedtuple
from copy import copy
from enum import Enum
from fractions import Fraction
from functools import lru_cache, wraps

from .tools import canonicalize


__all__ = [
    'Game',
]


def gamify_inputs(f):
    """
    A decorator to cast all arguments to a function into Games.

    Parameters
    ----------
    f : func
        The function to decorate.

    Returns
    -------
    inner : func
        The decorated function.
    """
    @wraps(f)
    def inner(*args):
        args = tuple(Game(_) for _ in args)
        return f(*args)

    return inner


class Outcome(Enum):
    """
    Representing the outcome classes of Games.
    """
    PREVIOUS = 'P'
    NEXT     = 'N'
    LEFT     = 'L'
    RIGHT    = 'R'


class Game(object):
    """
    A game consists of a *left set* and a *right set*, and is often written:

        { G_L | G_R }

    where G_L is the left set of Games, and G_R is the right set of Games. The trivial game, {|},
    is known as 0 and is said to have been born on day 0. Recursively, we can define games born
    on day `n` as all possible games whose left set and right set consist of Games having birthdays
    no later than day `n-1`.
    """

    def __init__(G, left=None, right=None):
        """
        Construct a Game.

        Parameters
        ----------
        left : iterable, number, Game, None
            If an iterable, it is the left set of the game. If a number, construct the game
            corresponding to that number. If a Game, use its left set, and if None, treat the left
            set as empty.
        right : iterable, None
            The right set of the game. If None, treat the right set as empty.
        """
        # passed in a number
        if isinstance(left, (int, float, Fraction)):
            v = float(left)
            if v.is_integer():
                if v == 0:
                    G._left = set()
                    G._right = set()
                elif v > 0:
                    G._left = {Game(v-1)}
                    G._right = set()
                else:
                    G._left = set()
                    G._right = {Game(v+1)}
            else:
                _, d = v.as_integer_ratio()
                if d > 1024:
                    msg = f"{v} would result in a very deep game tree."
                    raise ValueError(msg)
                G._left = {Game(v - 1/d)}
                G._right = {Game(v + 1/d)}
        # passed in a game
        elif isinstance(left, Game):
            G._left = {Game(_) for _ in left._left}
            G._right = {Game(_) for _ in left._right}
        # passed in sets
        else:
            if left is not None:
                G._left = {Game(_) for _ in left}
            else:
                G._left = set()
            if right is not None:
                G._right = {Game(_) for _ in right}
            else:
                G._right = set()

    @gamify_inputs
    # @lru_cache(maxsize=None)
    def __ge__(G, H):
        """
        A Game G is greater than or equal to a Game H so long as there are no G_R such that
        H >= G_R and there are no H_L such that H_L >= G.

        Parameters
        ----------
        H : Game
            The second game.

        Returns
        -------
        ge : bool
            G >= H.
        """
        a = all(not H >= G_R for G_R in G._right)
        b = all(not H_L >= G for H_L in H._left)
        return a and b

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __le__(G, H):
        """
        G <= H if H >= G.

        Parameters
        ----------
        H : Game
            The second game.

        Returns
        -------
        le : bool
            G <= H.
        """
        return H >= G

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __gt__(G, H):
        """
        G > H if G >= H and not G == H.

        Parameters
        ----------
        H : Game
            The second game.

        Returns
        -------
        gt : bool
            G > H.
        """
        return G >= H and not H >= G

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __lt__(G, H):
        """
        G < H if G <= H and not G == H.

        Parameters
        ----------
        H : Game
            The second game.

        Returns
        -------
        lt : bool
            G < H.
        """
        return H > G

    @gamify_inputs
    # @lru_cache(maxsize=None)
    def __eq__(G, H):
        """
        G == H if G >= H and G <= H.

        Parameters
        ----------
        H : Game
            The second game.

        Returns
        -------
        eq : bool
            G == H.
        """
        if hash(G) == hash(H):
            return True
        else:
            return G >= H and H >= G

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __or__(G, H):
        """
        G | H ('fuzzy' or 'confused with') if neither G >= H nor G <= H.

        Parameters
        ----------
        H : Game
            The second game.

        Returns
        -------
        fuzzy : bool
            G | H.
        """
        return not (G >= H or H >= G)

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __ror__(G, H):
        """
        G | H ('fuzzy' or 'confused with') if neither G >= H nor G <= H.

        Parameters
        ----------
        H : Game
            The second game.

        Returns
        -------
        fuzzy : bool
            G | H.
        """
        return not (G >= H or H >= G)

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __neg__(G):
        """
        -{G_L | G_R} is recursively defined as {-G_R | -G_L}.

        Returns
        -------
        ng : Game
            -G.
        """
        return Game({-G_R for G_R in G._right}, {-G_L for G_L in G._left})

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __add__(G, H):
        """
        The disjoint sum of two games:
            G + H = {G + H_L, H + G_L | G + H_R, H + G_R}

        Parameters
        ----------
        H : Game
            The second Game.

        Returns
        -------
        GplusH : Game
            The sum of G and H.
        """
        if not G.is_number and H.is_number:
            lefts = {G_L + H for G_L in G._left}
            rights = {G_R + H for G_R in G._right}
            return Game(lefts, rights)
        elif G.is_number and not H.is_number:
            lefts = {H_L + G for H_L in H._left}
            rights = {H_R + G for H_R in H._right}
            return Game(lefts, rights)
        else:
            left_a = {H + G_L for G_L in G._left}
            left_b = {G + H_L for H_L in H._left}
            right_a = {H + G_R for G_R in G._right}
            right_b = {G + H_R for H_R in H._right}
            return Game(left_a | left_b, right_a | right_b)

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __radd__(G, H):
        """
        The disjoint sum of two games:
            G + H = {G + H_L, H + G_L | G + H_R, H + G_R}

        Parameters
        ----------
        H : Game
            The second Game.

        Returns
        -------
        GplusH : Game
            The sum of G and H.
        """
        return G + H

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __sub__(G, H):
        """
        The disjoint difference of two games:
            G - H = G + (-H)

        Parameters
        ----------
        H : Game
            The second Game.

        Returns
        -------
        GminusH : Game
            The difference of G and H.
        """
        return G + (-H)

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __mul__(G, H):
        """
        The Norton product of G and H.

        Parameters
        ----------
        H : Game
            The second Game.

        Returns
        -------
        GtimesH : Game
            The product of G and H.
        """
        left_a = {G_L * H + G * H_L - G_L * H_L for G_L in G._left for H_L in H._left}
        left_b = {G_R * H + G * H_R - G_R * H_R for G_R in G._right for H_R in H._right}
        right_a = {G_L * H + G * H_R - G_L * H_R for G_L in G._left for H_R in H._right}
        right_b = {G_R * H + G * H_L - G_R * H_L for G_R in G._right for H_L in H._left}
        return Game(left_a | left_b, right_a | right_b)

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __rmul__(G, H):
        """
        The Norton product of G and H.

        Parameters
        ----------
        H : Game
            The second Game.

        Returns
        -------
        GtimesH : Game
            The product of G and H.
        """
        return G * H

    @gamify_inputs
    @lru_cache(maxsize=None)
    def _inverse(G):
        """
        Invert the Game G.

        Returns
        -------
        Ginv : Game
            The inverse of G.
        """
        # TODO: implement this.
        raise NotImplementedError

    @gamify_inputs
    @lru_cache(maxsize=None)
    def __truediv__(G, H):
        """
        The Norton quotient of G by H:

            G / H = G * (1/H)

        Parameters
        ----------
        H : Game
            The second Game.

        Returns
        -------
        GdivH : Game
            The quotient of G by H.
        """
        return G * H._inverse()

    def __hash__(G):
        """
        Define the hash of a game as the hash of the left and right options.

        Returns
        -------
        hash : str
            The hash of G.
        """
        return hash((frozenset(G._left), frozenset(G._right)))

    def __str__(G):
        """
        The value of G as a string.

        Returns
        -------
        s : str
            The value of G as a string.
        """
        return G.value

    def __repr__(G):
        """
        Construct a representation of G. It is of the form {G_L|G_R}, where the options are
        their shorthand values if possible.

        Returns
        -------
        r : str
            A representation of G.
        """
        lefts = ','.join(G_L.value for G_L in sorted(G._left, key=str))
        rights = ','.join(G_R.value for G_R in sorted(G._right, key=str))
        return '{{{}|{}}}'.format(lefts, rights)

    def subpositions(G):
        """
        Return an iterator over all subpositions of G.

        Yields
        ------
        g : Game
            A subposition of G.
        """
        for g in G._left | G._right:
            yield g
            yield from g.subpositions()

    @property
    def is_impartial(G):
        """
        A game is impartial if its left options equal its right options, and each of its options
        are impartial.

        Returns
        -------
        impartial : bool
            Whether G is impartial or not.
        """
        if G._left == G._right:
            if all(G_L.is_impartial for G_L in G._left):
                return True
        return False

    @property
    def is_dicotic(G):
        """
        A dicotic, or all-small, Game is one where either both or neither player have options
        at every subposition.

        Returns
        -------
        dicotic : bool
            Whether the Game is dicotic or not.
        """
        if not (bool(G._left) ^ bool(G._right)):
            if all(g.is_dicotic for g in G._left | G._right):
                return True
        return False

    @property
    def is_infinitesimal(G):
        """
        A game is infinitesimal if it is non-zero and smaller than any positive number and greater
        than any negative number. Equivalently, it's left and right stops are both zero. Note, this
        does not imply that an infinitesimal can not be positive (> 0) or negative (< 0).

        Returns
        -------
        infinitesimal : bool
            Whether the game is infinitesimal or not.
        """
        from .tools import left_stop, right_stop
        a = left_stop(G, adorn=False) == right_stop(G, adorn=False) == 0
        b = G != 0
        return a and b

    @property
    def is_number(G):
        """
        A game is a (surreal) number if each of it left options is less than each of its right
        options, and all its options are also numbers.

        Returns
        -------
        number : bool
            Whether G is a number or not.

        Note
        ----
        The definition implies that, in canonical form, the number has at most one left option and
        one right option, because one must dominate. Also, if either of the two sets of options is
        empty, the number is an integer.
        """
        a = all(g.is_number for g in G._left | G._right)
        b = all(G_L < G_R for G_L in G._left for G_R in G._right)
        return a and b

    @property
    def is_numberish(G):
        """
        A Game is numberish if it is infinitesimally close to a number.

        Returns
        -------
        numberish : bool
            Whether G is numberish or not.
        """
        from .tools import left_stop, right_stop
        return left_stop(G, adorn=False) == right_stop(G, adorn=False)

    @property
    def is_switch(G):
        """
        A game is a switch if it's left and right options are numbers, and G_L > G_R.

        Returns
        -------
        number : bool
            Whether G is a switch or not.
        """
        if len(G._left) == len(G._right) == 1:
            G_L = next(iter(G._left))
            G_R = next(iter(G._right))
            if G_L.is_number and G_R.is_number and G_L > G_R:
                return True
        return False

    @property
    def birthday(G):
        """
        The birthday of a game G is defined as the one more than the maximum birthday of its
        options.

        Returns
        -------
        bday : int
            G's birthday.
        """
        if G._left == G._right == set():
            bday = 0
        else:
            bday = max(g.birthday for g in G._left | G._right) + 1

        return bday

    @property
    def outcome(G):
        """
        The outcome of a Game belongs to one of four equivalence classes:

            P : A win for the second player.
            N : A win for the first player.
            L : A win for the left player, regardless of play order.
            R : A win for the right player, regardless of play order.

        Returns
        -------
        o : Outcome
            The outcome class of game G.
        """
        zero = Game()
        if G == zero:
            return Outcome.PREVIOUS
        elif G > zero:
            return Outcome.LEFT
        elif G < zero:
            return Outcome.RIGHT
        else:
            return Outcome.NEXT

    @property
    def value(G):
        """
        The value of the game in short-hand.

        Returns
        -------
        v : str
            The value of the Game as a string.
        """
        return _value_str(G)


Symbol = namedtuple('Symbol', ['unicode', 'latex'])


_SYMBOLS = {
    'star': Symbol('*', r'\ensuremath{\ast}'),
    'up': Symbol('↑', r'\ensuremath{\uparrow}'),
    'down': Symbol('↓', r'\ensuremath{\downarrow}'),
    '2up': Symbol('↑↑', r'\ensuremath{\Uparrow}'),
    '2down': Symbol('↓↓', r'\ensuremath{\Downarrow}'),
    'cdot': Symbol('·', r'\ensuremath{\cdot}'),
    'pm': Symbol('±', r'\ensuremath{\pm}'),
    'frac': Symbol((lambda n, d: f"{n}/{d}"), (lambda n, d: f'\\ensuremath{{\\nicefrac{{{n}}}{{{d}}}}}')),
    'miny': Symbol((lambda g: '⧿_' + g), (lambda g: r'\ensuremath{\tminus}_{' + g + '}')),
    'tiny': Symbol((lambda g: '⧾_' + g), (lambda g: r'\ensuremath{\tplus}_{' + g + '}')),
}


@lru_cache(maxsize=None)
def _value_str(G, latex=False):
    """
    A string representation of the game, accounting for named special games.

    Parameters
    ----------
    G : Game
        The game to compute the representation of.

    Returns
    -------
    s : str
        The name of the game.
    """
    G = canonicalize(G)
    G_hash = hash(G)

    zero = Game()

    if not len(G._left | G._right):
        return '0'

    star = Game({zero}, {zero})

    if G_hash == hash(star):
        return _SYMBOLS['star'][bool(latex)]

    up = hash(Game({zero}, {star}))
    up_star = hash(Game({zero, star}, {zero}))
    down = hash(Game({star}, {zero}))
    down_star = hash(Game({zero}, {zero, star}))

    if G_hash == up:
        return _SYMBOLS['up'][bool(latex)]
    elif G_hash == up_star:
        return _SYMBOLS['up'][bool(latex)] + _SYMBOLS['star'][bool(latex)]
    elif G_hash == down:
        return _SYMBOLS['down'][bool(latex)]
    elif G_hash == down_star:
        return _SYMBOLS['down'][bool(latex)] + _SYMBOLS['star'][bool(latex)]

    g = copy(G)
    i = 1
    while True:
        if g._left == {zero} and len(g._right) == 1:
            i += 1
            g = next(iter(g._right))
            g_hash = hash(g)
            if g_hash == up:
                return f"{i}" + _SYMBOLS['cdot'][bool(latex)] + \
                                _SYMBOLS['up'][bool(latex)] + \
                                _SYMBOLS['star'][bool(latex)] * ((i+1) % 2)
            elif g_hash == up_star:
                return f"{i}" + _SYMBOLS['cdot'][bool(latex)] + \
                                _SYMBOLS['up'][bool(latex)] + \
                                _SYMBOLS['star'][bool(latex)] * (i % 2)
        else:
            break

    g = copy(G)
    i = 1
    while True:
        if g._right == {zero} and len(g._left) == 1:
            i += 1
            g = next(iter(g._left))
            g_hash = hash(g)
            if g_hash == down:
                return f"{i}" + _SYMBOLS['cdot'][bool(latex)] + \
                                _SYMBOLS['down'][bool(latex)] + \
                                _SYMBOLS['star'][bool(latex)] * ((i+1) % 2)
            elif g_hash == down_star:
                return f"{i}" + _SYMBOLS['cdot'][bool(latex)] + \
                                _SYMBOLS['down'][bool(latex)] + \
                                _SYMBOLS['star'][bool(latex)] * (i % 2)
        else:
            break

    # nimbers
    if G.is_impartial:
        return _SYMBOLS['star'][bool(latex)] + f"{len(G._left)}"

    if G.is_number:
        # non-zero integer
        if G._left:
            lf = Fraction(next(iter(G._left)).value)
        if G._right:
            rf = Fraction(next(iter(G._right)).value)

        if not G._right:
            return str(lf + 1)
        if not G._left:
            return str(rf - 1)

        v = (lf + rf) / 2
        n, d = v.numerator, v.denominator
        return _SYMBOLS['frac'][bool(latex)](n, d)

    if len(G._left) == len(G._right) == 1:
        G_L, G_R = next(iter(G._left)), next(iter(G._right))
        if G_L.is_number and G_R.is_number:
            # switches
            if G_L > G_R:
                lf = Fraction(G_L.value)
                rf = Fraction(G_R.value)
                mean = (lf + rf) / 2
                diff = (lf - rf) / 2
                if mean:
                    return f"{mean}" + _SYMBOLS['pm'][bool(latex)] + f"{diff}"
                else:
                    return _SYMBOLS['pm'][bool(latex)] + f"{diff}"

            # tepid games
            else:  # G_L == G_R; G_L < G_R caught by numbers above.
                return f"{G_L.value}" + _SYMBOLS['star'][bool(latex)]

    if G._left == {zero} and len(G._right) == 1:
        G_R = next(iter(G._right))
        if G_R._left == {zero} and len(G_R._right) == 1:
            G_RR = next(iter(G_R._right))
            return _SYMBOLS['tiny'][bool(latex)](f"{(-G_RR).value}")

    if G._right == {zero} and len(G._left) == 1:
        G_L = next(iter(G._left))
        if G_L._right == {zero} and len(G_L._left) == 1:
            G_LL = next(iter(G_L._left))
            return _SYMBOLS['miny'][bool(latex)](f"{(G_LL).value}")

    #todo: sums

    return repr(G)



"""
Tests for ludology.nimber.
"""

import pytest

from ludology import Game, Nimber, FarStar
from ludology.closet import one


def test_add_nimbers():
    """
    Test that Nimbers add correctly.
    """
    a = Nimber(5)
    b = Nimber(6)
    c = Nimber(3)
    assert a + b == c


def test_add_nimber_game():
    """
    Test that Nimbers add with Games correctly.
    """
    a = Nimber(3)
    b = Game(Nimber(2))
    c = Game(Nimber(1))
    assert a + b == c


def test_mul_nimbers():
    """
    Test that Nimbers multiply correctly.
    """
    a = Nimber(2)
    b = Nimber(6)
    c = Nimber(11)
    assert a * b == c


def test_mul_nimber_game():
    """
    Test that Nimbers multiply with Games correctly.
    """
    a = Nimber(2)
    b = Game(Nimber(2))
    c = Game(Nimber(3))
    assert a * b == c


def test_neg_nimbers():
    """
    Test that a Nimber is its own negative.
    """
    a = Nimber(5)
    assert a == -a


def test_eq_nimber_game():
    """
    Test that Nimbers and Games equate properly.
    """
    a = Nimber(3)
    b = 3 * one
    assert not a == b


def test_fail():
    """
    Test that Nimbers must have integral values.
    """
    with pytest.raises(ValueError):
        Nimber(0.5)


@pytest.mark.parametrize(['n', 'v'], [
    (0, '0'),
    (1, '*'),
    (6, '*6'),
])
def test_nimber_value(n, v):
    """
    Test that the value of a Nimber is represented correctly.
    """
    assert Nimber(n).value == v


@pytest.mark.parametrize('n', [
    Nimber(3),
    Nimber(4),
    Nimber(7),
])
def test_nimber_is_number(n):
    """
    Test that Nimbers (other than 0) are not numbers.
    """
    assert not n.is_number


def test_nimber_is_number_2():
    """
    Test that the Nimber 0 is a number.
    """
    assert Nimber(0).is_number


@pytest.mark.parametrize('n', [
    Nimber(3),
    Nimber(4),
    Nimber(7),
])
def test_nimber_is_infinitesimal(n):
    """
    Test that Nimbers (other than 0) are infinitesimal.
    """
    assert n.is_infinitesimal


def test_nimber_is_infinitesimal_2():
    """
    Test that the Nimber 0 is not infinitesimal.
    """
    assert not Nimber(0).is_infinitesimal


@pytest.mark.parametrize('n', [
    Nimber(0),
    Nimber(3),
    Nimber(7),
])
def test_nimber_is_impartial(n):
    """
    Test that all Nimbers are impartial.
    """
    assert n.is_impartial


@pytest.mark.parametrize('n', [
    Nimber(0),
    Nimber(3),
    Nimber(7),
])
def test_nimber_is_numberish(n):
    """
    Test that all Nimbers are numberish.
    """
    assert n.is_numberish


@pytest.mark.parametrize('n', [
    Nimber(0),
    Nimber(3),
    Nimber(7),
])
def test_nimber_is_dicotic(n):
    """
    Test that all Nimbers are dicotic.
    """
    assert n.is_dicotic


@pytest.mark.parametrize('n', [
    0,
    3,
    7,
])
def test_nimber_birthday(n):
    """
    Test that Nimbers have the right birthdays.
    """
    assert Nimber(n).birthday == n


@pytest.mark.parametrize('n', [
    Nimber(0),
    Nimber(1),
    Nimber(2),
])
def test_far_star_sum(n):
    """
    Test that far-star sums to far-star.
    """
    fs = FarStar()
    assert fs + n == fs


def test_far_star_value():
    """
    Test that far star has the right value.
    """
    assert FarStar().value == "☆"


@pytest.mark.parametrize('n', [
    Nimber(0),
    Nimber(2),
    Nimber(7),
])
def test_far_star_left(n):
    """
    Test that far star's left set contains an arbitrary Nimber.
    """
    assert n in FarStar()._left


@pytest.mark.parametrize('n', [
    Nimber(0),
    Nimber(2),
    Nimber(7),
])
def test_far_star_right(n):
    """
    Test that far star's right set contains an arbitrary Nimber.
    """
    assert n in FarStar()._right



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ludology']

package_data = \
{'': ['*']}

install_requires = \
['numpy', 'scipy']

extras_require = \
{'dev': ['codecov',
         'hypothesis',
         'ipython[nbconvert]',
         'pytest',
         'pytest-cov',
         'sphinx'],
 'plotting': ['matplotlib']}

setup(name='ludology',
      version='0.1.1',
      description='Ludology is a package for the study of combinatorial game theory.',
      author='Ryan G. James',
      author_email='ryangregoryjames@gmail.com',
      url='https://github.com/autoplectic/ludology',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )
